"""Exceptions for event processor."""


class EventProcessorError(BaseException):
    """General exception for the event-processor library."""


class FilterError(EventProcessorError):
    """Exception for failures related to filters."""


class InvocationError(EventProcessorError):
    """Exception for failures in invocation."""



from unittest.mock import Mock

import pytest

from src.event_processor.invocation_strategies import InvocationStrategies
from src.event_processor.dependencies import Event, Depends
from src.event_processor.event_processor import EventProcessor
from src.event_processor.filters import Eq, Exists, Accept, Dyn


@pytest.fixture
def event_processor():
    return EventProcessor()


def test_simple_filtering(event_processor):
    @event_processor.processor(Eq("top.mid.low", "val"))
    def my_processor(event: Event):
        return event["top"]["mid"]["low"]

    result = event_processor.invoke({"top": {"mid": {"low": "val", "other": 1234}}})

    assert result == "val"


@pytest.mark.parametrize("test_event, ret", [({"key": "val"}, "val"), ({"key": 123}, 123), ({"key": None}, None)])
def test_any_filter(test_event, ret, event_processor):
    @event_processor.processor(Exists("key"))
    def my_processor(event: Event):
        return event["key"]

    result = event_processor.invoke(test_event)

    assert result == ret


def test_simple_pre_processing(event_processor):
    def pre_processor(event: Event):
        return event["user"]

    @event_processor.processor(Exists("user"))
    def my_processor(user_dict=Depends(pre_processor)):
        return user_dict

    result = event_processor.invoke({"user": {"name": "John"}})

    assert result == {"name": "John"}


def test_processor_with_dependencies(event_processor):
    dependency_mock = Mock()

    def get_dependency():
        return dependency_mock

    @event_processor.processor(Accept())
    def my_processor(_event: Event, dep: Mock = Depends(get_dependency)):
        return dep

    result = event_processor.invoke({})

    assert result is dependency_mock


def test_pre_processor_with_dependencies(event_processor):
    dependency_mock = Mock()

    def dep_factory():
        return dependency_mock

    def pre_processor(_event: Event, dep=Depends(dep_factory)):
        return dep

    @event_processor.processor(Accept())
    def my_processor(dep_from_pre_processor=Depends(pre_processor)):
        return dep_from_pre_processor

    result = event_processor.invoke({})

    assert result is dependency_mock


def test_ambiguous_filters_with_rank(event_processor):
    mock_a = Mock()
    mock_b = Mock()

    @event_processor.processor(Exists("a"))
    def a_processor():
        mock_a()

    @event_processor.processor(Eq("a", "b"), rank=1)
    def b_processor():
        mock_b()

    event_processor.invoke({"a": "b"})

    mock_a.assert_not_called()
    mock_b.assert_called_once()


def test_ambiguous_filters_with_no_rank(event_processor):
    mock_a = Mock()
    mock_b = Mock()

    @event_processor.processor(Exists("a"))
    def a_processor():
        mock_a()

    @event_processor.processor(Eq("a", "b"))
    def b_processor():
        mock_b()

    event_processor.invoke({"a": "b"})

    mock_a.assert_called_once()
    mock_b.assert_not_called()


def test_ambiguous_filters_with_no_rank_and_non_default_invocation_strategy():
    event_processor = EventProcessor(invocation_strategy=InvocationStrategies.NO_MATCHES)
    mock_a = Mock()
    mock_b = Mock()

    @event_processor.processor(Exists("a"))
    def a_processor():
        mock_a()

    @event_processor.processor(Eq("a", "b"))
    def b_processor():
        mock_b()

    event_processor.invoke({"a": "b"})

    mock_a.assert_not_called()
    mock_b.assert_not_called()


def test_dynamic_filters_with_lambda_function(event_processor):
    filter_a = Dyn(lambda e: e["messages"][0].get("process") is True)
    mock_a = Mock()
    mock_b = Mock()

    @event_processor.processor(Accept(), rank=-1)
    def default_processor():
        mock_b()

    @event_processor.processor(filter_a)
    def a_processor():
        mock_a()

    event_processor.invoke({"messages": [{"process": True}]})
    event_processor.invoke({"messages": [{"process": False}]})

    mock_a.assert_called_once()
    mock_b.assert_called_once()


def test_dynamic_filters_with_dedicated_function(event_processor):
    mock_a = Mock()
    mock_b = Mock()

    def some_dependency():
        return 0

    def my_filter(event: Event, dep_value: int = Depends(some_dependency)):
        return event["key"] == dep_value

    @event_processor.processor(Dyn(my_filter))
    def my_processor():
        mock_a()

    @event_processor.processor(Accept(), rank=-1)
    def default_processor():
        mock_b()

    event_processor.invoke({"key": 0})
    event_processor.invoke({"key": 1})

    mock_a.assert_called_once()
    mock_b.assert_called_once()



from typing import Any


def get_value_at_path(source: dict, path: str) -> Any:
    """Get the value in a source dict at the given path.

    :param source: The dict from which to extract the value
    :param path: The path to look into
    :return: The value found at the path
    :raises: KeyError when no value exists at the path
    """
    current_location = source
    for part in path.split("."):
        current_location = current_location[part]

    return current_location



from . import filters  # noqa
from .dependencies import Depends, Event  # noqa
from .event_processor import EventProcessor  # noqa
from .exceptions import (  # noqa
    EventProcessorError,
    FilterError,
    InvocationError,
)
from .invocation_strategies import InvocationStrategies  # noqa



from unittest.mock import Mock

import pytest
from pydantic import BaseModel

from src.event_processor.dependencies import Depends, Event
from src.event_processor.event_processor import EventProcessor, processor_params_are_valid
from src.event_processor.exceptions import (
    FilterError,
    InvocationError,
)
from src.event_processor.filters import Exists, Accept, Eq

MOD_PATH = "src.event_processor.event_processor"


@pytest.fixture
def event_processor():
    return EventProcessor()


def test_add_subprocessor_adds_subprocessor(event_processor):
    other_event_processor = EventProcessor()
    other_event_processor.processors[Accept(), 0] = lambda: 0

    event_processor.add_subprocessor(other_event_processor)

    assert len(event_processor.processors) == 1


def test_add_subprocessor_raises_for_existing_filters(event_processor):
    other_event_processor = EventProcessor()
    other_event_processor.processors[Accept(), 0] = lambda: 0
    event_processor.processors[Accept(), 0] = lambda: 1

    with pytest.raises(FilterError):
        event_processor.add_subprocessor(other_event_processor)


def test_processor_registers_a_processor(event_processor):
    filter_ = Exists("a")

    @event_processor.processor(filter_)
    def a_test():
        pass

    assert len(event_processor.processors) == 1


def test_processor_raises_exception_when_filter_exists(event_processor):
    filter_ = Exists("a")
    event_processor.processors[filter_, 0] = 0

    with pytest.raises(FilterError):

        @event_processor.processor(filter_)
        def a_test():
            pass


def test_processor_raises_exception_when_the_processor_takes_invalid_params(event_processor):

    with pytest.raises(FilterError):

        @event_processor.processor(Accept())
        def fn(_x):
            pass


def test_invoke_calls_matching_processor(event_processor):
    filter_ = Exists("a")
    called = False

    @event_processor.processor(filter_)
    def a_test():
        nonlocal called
        called = True

    event_processor.invoke({"a": 0})
    assert called is True


def test_invoke_raises_for_no_matching_processors(event_processor):

    with pytest.raises(InvocationError):
        event_processor.invoke({"a": 0})


def test_invoke_returns_the_processor_return_value(event_processor):
    result_mock = Mock()
    event_processor.processors[Accept(), 0] = lambda: result_mock

    result = event_processor.invoke({"a": 0})

    assert result is result_mock


def test_invoke_injects_event_into_processor(event_processor):
    event = {"a": 0}
    received_event = False

    @event_processor.processor(Accept())
    def fn(ev: Event):
        nonlocal received_event
        received_event = ev == event

    event_processor.invoke(event)

    assert received_event is True


def test_invoke_injects_dependencies_into_processor(event_processor):
    event = {"a": 0}
    dependency_result = Mock()

    @event_processor.processor(Accept())
    def fn(dep: Mock = Depends(lambda: dependency_result)):
        dep.method()

    event_processor.invoke(event)

    dependency_result.method.assert_called_once()


def test_invoke_calls_highest_ranking_processor(event_processor):
    event = {"a": 0}
    called_a = False
    called_b = False

    @event_processor.processor(Exists("a"))
    def fn_a():
        nonlocal called_a
        called_a = True

    @event_processor.processor(Eq("a", 0), rank=1)
    def fn_b():
        nonlocal called_b
        called_b = True

    event_processor.invoke(event)

    assert called_a is False
    assert called_b is True


def test_invoke_calls_negative_rank_as_fallback(event_processor):
    called_a = False
    called_b = False

    @event_processor.processor(Accept(), rank=-1)
    def fn_a():
        nonlocal called_a
        called_a = True

    @event_processor.processor(Exists("a"))
    def fn_b():
        nonlocal called_b
        called_b = True

    event_processor.invoke({"not-a": 0})

    assert called_a is True
    assert called_b is False


def test_processor_params_are_valid_returns_true_for_valid_params():
    def processor(_a: Event, _b: Event, _c: BaseModel, _d=Depends(Mock())):
        pass

    assert processor_params_are_valid(processor) is True


def test_processor_params_are_valid_returns_false_for_invalid_params():
    def processor(_a: Event, _b: Event, _c: BaseModel, _d=Depends(Mock()), _e=0):
        pass

    assert processor_params_are_valid(processor) is False



from unittest.mock import Mock, patch

import pytest

from src.event_processor.dependencies import Event
from src.event_processor.exceptions import FilterError
from src.event_processor.filters import Exists, Accept, Eq, And, Or, Lt, NumCmp, Leq, Gt, Geq, Dyn


def test_filter_and_creates_and_filter():
    a_filter = Exists("a")
    b_filter = Exists("b")
    c_filter = Exists("c")

    combined = a_filter & b_filter & c_filter

    assert isinstance(combined, And)
    assert combined.filters[0].filters[0] is a_filter
    assert combined.filters[0].filters[1] is b_filter
    assert combined.filters[1] is c_filter


def test_filter_or_creates_or_filter():
    a_filter = Exists("a")
    b_filter = Exists("b")
    c_filter = Exists("c")

    combined = a_filter | b_filter | c_filter

    assert isinstance(combined, Or)
    assert combined.filters[0].filters[0] is a_filter
    assert combined.filters[0].filters[1] is b_filter
    assert combined.filters[1] is c_filter


def test_filter_operators_when_combined_respect_priority_of_operations_and_commutativity():
    a_filter = Exists("a")
    b_filter = Exists("b")
    c_filter = Exists("c")

    combined_1 = a_filter & (b_filter | c_filter)
    combined_1_commutativity = (b_filter | c_filter) & a_filter
    combined_2 = a_filter | (b_filter & c_filter)
    combined_2_commutativity = (b_filter & c_filter) | a_filter

    assert combined_1.matches({"a": 0, "c": 0})
    assert combined_1_commutativity.matches({"a": 0, "c": 0})
    assert combined_2.matches({"a": 0})
    assert combined_2_commutativity.matches({"a": 0})


@pytest.mark.parametrize("event", [{}, {"key": "value"}, {"top": {"middle": {"lower": ""}}}, None])
def test_accept_filter_matches_any_event(event):
    result = Accept().matches(event)

    assert result is True


def test_accept_filter_is_equal_to_other_accept_filters():
    a_filter = Accept()
    b_filter = Accept()

    assert a_filter == b_filter


def test_accept_filter_is_not_equal_to_other_object():
    a_filter = Accept()
    b_filter = 3

    assert a_filter != b_filter


def test_accept_filter_hash_returns_class_hash():
    filter_ = Accept()

    assert hash(filter_) == hash(filter_.__class__)


def test_exists_filter_matches_event_with_existing_top_level_value():
    test_filter = Exists("top-level")

    result = test_filter.matches({"top-level": "value"})

    assert result is True


def test_exists_filter_matches_event_with_existing_nested_value():
    test_filter = Exists("top.mid.low")

    result = test_filter.matches({"top": {"mid": {"low": "val"}}})

    assert result is True


def test_exists_filter_does_not_match_empty_event():
    test_filter = Exists("key")

    result = test_filter.matches({})

    assert result is False


def test_exists_filter_is_equal_to_other_exists_filter_with_the_same_path():
    a_filter = Exists("a")
    b_filter = Exists("a")

    assert a_filter == b_filter


def test_exists_filter_is_not_equal_to_other_types():
    a_filter = Exists("a")
    b_filter = 0

    assert a_filter != b_filter


def test_exists_filter_is_not_equal_to_different_path():
    a_filter = Exists("a")
    b_filter = Exists("b")

    assert a_filter != b_filter


def test_exists_filter_hash_is_path_hash():
    filter_ = Exists("a")

    assert hash(filter_) == hash((Exists, "a"))


def test_eq_filter_matches_equal_top_level_value():
    test_filter = Eq("top-level", "expected-value")

    result = test_filter.matches({"top-level": "expected-value"})

    assert result is True


def test_eq_filter_matches_equal_nested_value():
    test_filter = Eq("top.mid.low", "expected-value")

    result = test_filter.matches({"top": {"mid": {"low": "expected-value"}}})

    assert result is True


def test_eq_filter_does_not_match_not_equal_values():
    test_filter = Eq("top", "value")

    result = test_filter.matches({"top": "not-value"})

    assert result is False


def test_eq_filter_does_not_match_nonexistent_key():
    test_filter = Eq("key", "value")

    result = test_filter.matches({})

    assert result is False


def test_eq_filter_is_equal_to_other_exists_filter_with_the_same_path_and_same_value():
    a_filter = Eq("a", 0)
    b_filter = Eq("a", 0)

    assert a_filter == b_filter


def test_eq_filter_is_not_equal_to_other_types():
    a_filter = Eq("a", 0)
    b_filter = 0

    assert a_filter != b_filter


def test_eq_filter_is_not_equal_to_different_path():
    a_filter = Eq("a", 0)
    b_filter = Eq("b", 0)

    assert a_filter != b_filter


def test_eq_filter_is_not_equal_to_different_value():
    a_filter = Eq("a", 0)
    b_filter = Eq("a", 1)

    assert a_filter != b_filter


def test_eq_filter_hash_hashes_path_and_value():
    filter_ = Eq("a", 0)

    assert hash(filter_) == hash((Eq, ("a", 0)))


def test_num_cmp_filter_matches_when_comparator_returns_true():
    comparator_mock = Mock(return_value=True)
    filter_ = NumCmp("a", comparator_mock, 0)

    result = filter_.matches({"a": 0})

    assert result is True


def test_num_cmp_filter_does_not_match_when_comparator_returns_false():
    comparator_mock = Mock(return_value=False)
    filter_ = NumCmp("a", comparator_mock, 0)

    result = filter_.matches({"a": 0})

    assert result is False


def test_num_cmp_filter_hash_hashes_path_and_comparator():
    mock_comparator = Mock()
    assert hash(NumCmp("a", mock_comparator, 0)) == hash((NumCmp, ("a", mock_comparator, 0)))


def test_num_cmp_filter_is_equal_when_path_and_comparator_and_target_are_equal():
    mock_comparator = Mock()
    a_cmp = NumCmp("a", mock_comparator, 0)
    b_cmp = NumCmp("a", mock_comparator, 0)

    assert a_cmp == b_cmp


def test_num_cmp_filter_is_not_equal_when_path_or_comparator_or_target_are_different():
    mock_comparator = Mock()
    other_mock_comparator = Mock()
    a_cmp = NumCmp("a", mock_comparator, 0)
    b_cmp = NumCmp("b", mock_comparator, 0)
    c_cmp = NumCmp("a", other_mock_comparator, 0)
    d_cmp = NumCmp("a", mock_comparator, 1)

    assert not a_cmp == b_cmp
    assert not a_cmp == c_cmp
    assert not a_cmp == d_cmp


def test_num_cmp_filter_is_not_equal_when_comparing_with_another_type():
    mock_comparator = Mock()
    a_cmp = NumCmp("a", mock_comparator, 0)

    assert not a_cmp == 0


def test_lt_filter_matches_when_value_is_less_than():
    filter_ = Lt("a", 0)

    result = filter_.matches({"a": "-1"})

    assert result is True


def test_lt_filter_does_not_match_when_value_is_equal():
    filter_ = Lt("a", 0)

    result = filter_.matches({"a": "0"})

    assert result is False


def test_lt_filter_does_not_match_when_value_greater_than():
    filter_ = Lt("a", 0)

    result = filter_.matches({"a": "1"})

    assert result is False


def test_lt_filter_raises_value_error_when_input_value_is_not_a_float():
    with pytest.raises(ValueError):
        Lt("a", "not-a-float")


def test_lt_filter_does_not_match_when_path_does_not_exist():
    filter_ = Lt("a", 0)

    result = filter_.matches({"not-a": "-1"})

    assert result is False


def test_lt_filter_does_not_match_when_value_is_not_a_float():
    filter_ = Lt("a", 0)

    result = filter_.matches({"a": "not-a-float"})

    assert result is False


def test_leq_filter_matches_when_value_is_less_than():
    filter_ = Leq("a", 0)

    result = filter_.matches({"a": "-1"})

    assert result is True


def test_leq_filter_matches_when_value_is_equal():
    filter_ = Leq("a", 0)

    result = filter_.matches({"a": "0"})

    assert result is True


def test_leq_filter_does_not_match_when_value_greater_than():
    filter_ = Leq("a", 0)

    result = filter_.matches({"a": "1"})

    assert result is False


def test_leq_filter_raises_value_error_when_input_value_is_not_a_float():
    with pytest.raises(ValueError):
        Leq("a", "not-a-float")


def test_leq_filter_does_not_match_when_path_does_not_exist():
    filter_ = Leq("a", 0)

    result = filter_.matches({"not-a": "-1"})

    assert result is False


def test_leq_filter_does_not_match_when_value_is_not_a_float():
    filter_ = Leq("a", 0)

    result = filter_.matches({"a": "not-a-float"})

    assert result is False


def test_gt_filter_matches_when_value_is_greater_than():
    filter_ = Gt("a", 0)

    result = filter_.matches({"a": "1"})

    assert result is True


def test_gt_filter_does_not_match_when_value_is_equal():
    filter_ = Gt("a", 0)

    result = filter_.matches({"a": "0"})

    assert result is False


def test_gt_filter_does_not_match_when_value_is_less_than():
    filter_ = Gt("a", 1)

    result = filter_.matches({"a": "0"})

    assert result is False


def test_gt_filter_raises_value_error_when_input_value_is_not_a_float():
    with pytest.raises(ValueError):
        Gt("a", "not-a-float")


def test_gt_filter_does_not_match_when_path_does_not_exist():
    filter_ = Gt("a", 0)

    result = filter_.matches({"not-a": "-1"})

    assert result is False


def test_gt_filter_does_not_match_when_value_is_not_a_float():
    filter_ = Gt("a", 0)

    result = filter_.matches({"a": "not-a-float"})

    assert result is False


def test_geq_filter_matches_when_value_is_greater_than():
    filter_ = Geq("a", 0)

    result = filter_.matches({"a": "1"})

    assert result is True


def test_geq_filter_matches_when_value_is_equal():
    filter_ = Geq("a", 0)

    result = filter_.matches({"a": "0"})

    assert result is True


def test_geq_filter_does_not_match_when_value_is_less_than():
    filter_ = Geq("a", 1)

    result = filter_.matches({"a": "0"})

    assert result is False


def test_geq_filter_raises_value_error_when_input_value_is_not_a_float():
    with pytest.raises(ValueError):
        Geq("a", "not-a-float")


def test_geq_filter_does_not_match_when_path_does_not_exist():
    filter_ = Geq("a", 0)

    result = filter_.matches({"not-a": "-1"})

    assert result is False


def test_geq_filter_does_not_match_when_value_is_not_a_float():
    filter_ = Geq("a", 0)

    result = filter_.matches({"a": "not-a-float"})

    assert result is False


def test_dyn_filter_raises_when_resolver_is_not_callable():
    with pytest.raises(FilterError):
        Dyn("not-callable")


def test_dyn_filter_matches_when_callable_returns_true():
    mock_callable = Mock(return_value=True)
    filter_ = Dyn(mock_callable)

    result = filter_.matches({})

    assert result is True


def test_dyn_filter_matches_when_callable_returns_value():
    mock_callable = Mock(return_value=False)
    filter_ = Dyn(mock_callable)

    result = filter_.matches({})

    assert result is False


def test_dyn_filter_injects_event_when_callable_is_a_lambda():
    mock_event = Mock()
    filter_ = Dyn(lambda e: e is mock_event)

    result = filter_.matches(mock_event)

    assert result is True


@patch("src.event_processor.filters.call_with_injection")
def test_dyn_filter_calls_with_injection_when_callable_is_not_a_lambda(call_with_injection_mock):
    def dyn_filter():
        pass

    filter_ = Dyn(dyn_filter)

    filter_.matches({})
    call_with_injection_mock.assert_called_once_with(dyn_filter, event=Event({}), cache={})


def test_dyn_filter_hash_is_resolver_hash():
    def dyn_filter():
        pass

    filter_ = Dyn(dyn_filter)

    assert hash(filter_) == hash(dyn_filter)


def test_eq_filter_matches_when_resolvers_are_equal():
    mock_resolver = Mock()

    filter_a = Dyn(mock_resolver)
    filter_b = Dyn(mock_resolver)

    assert filter_a == filter_b


def test_eq_filter_does_not_match_when_resolvers_are_different():
    filter_a = Dyn(lambda: 0)
    filter_b = Dyn(lambda: 0)

    assert not (filter_a == filter_b)


def test_eq_filter_does_not_match_when_other_is_not_a_dyn_filter():
    mock_resolver = Mock()
    filter_ = Dyn(mock_resolver)

    assert not (filter_ == "x")


def test_and_filter_matches_when_all_filters_match():
    test_filter = And(Exists("a"), Exists("b"), Eq("c", "d"))

    result = test_filter.matches({"a": 0, "b": 0, "c": "d"})

    assert result is True


def test_and_filter_matches_when_all_nested_filters_match():
    test_filter = And(And(Exists("a"), Exists("b")), Eq("c", "d"))

    result = test_filter.matches({"a": 0, "b": 0, "c": "d"})

    assert result is True


def test_and_filter_does_not_match_when_one_filter_does_not_match():
    test_filter = And(Exists("a"), Exists("b"))

    result = test_filter.matches({"a": 0})

    assert result is False


def test_and_filter_is_short_circuiting_when_one_sub_filter_does_not_match():
    test_filter = And(Exists("a"), Exists("b"), Dyn(lambda e: e["nonexistent"]))
    test_filter_shorthand = Exists("a") & Exists("b") & Dyn(lambda e: e["nonexistent"])

    result = test_filter.matches({"a": 0})
    result_shorthand = test_filter_shorthand.matches({"a": 0})

    assert result is False
    assert result_shorthand is False


def test_and_filter_is_equal_when_sub_filters_are_equal():
    a_filter = Mock()
    a_filter.__eq__ = lambda _, x: x is a_filter
    b_filter = Mock()
    b_filter.__eq__ = lambda _, x: x is b_filter

    and_filter_1 = And(a_filter, b_filter)
    and_filter_2 = And(b_filter, a_filter)

    assert and_filter_1 == and_filter_2


def test_and_filter_is_not_equal_when_sub_filters_are_different():
    a_filter = Mock()
    a_filter.__eq__ = lambda _, x: x is a_filter
    b_filter = Mock()
    b_filter.__eq__ = lambda _, x: x is b_filter

    and_filter_1 = And(a_filter, b_filter)
    and_filter_2 = And(b_filter)
    and_filter_3 = And(a_filter)

    assert and_filter_1 != and_filter_2
    assert and_filter_2 != and_filter_3


def test_and_filter_is_not_equal_when_other_is_not_a_filter():
    a_filter = And(Mock())

    assert a_filter != 0


def test_and_filter_hash_is_hash_of_all_sub_filters():
    a_filter = Exists("a")
    b_filter = Exists("b")

    and_hash = hash(And(a_filter, b_filter))

    assert and_hash == hash((And, (a_filter, b_filter)))


def test_or_filter_matches_when_one_filter_matches():
    test_filter = Or(Exists("a"), Eq("", "y"), Exists("0"))

    result = test_filter.matches({"": "y"})

    assert result is True


def test_or_filter_matches_when_one_nested_filter_matches():
    test_filter = Or(Or(Exists("a"), Eq("", "y")), Exists("0"))

    result = test_filter.matches({"": "y"})

    assert result is True


def test_or_filter_does_not_match_when_no_filters_match():
    test_filter = Or(Exists("a"), Eq("x", "y"))

    result = test_filter.matches({"x": "not-y"})

    assert result is False


def test_or_filter_is_short_circuiting_when_one_filter_matches():
    test_filter = Or(Exists("a"), Exists("b"), Dyn(lambda e: e["nonexistent"]))
    test_filter_shorthand = Exists("a") | Exists("b") | Dyn(lambda e: e["nonexistent"])

    result = test_filter.matches({"b": 0})
    result_shorthand = test_filter_shorthand.matches({"b": 0})

    assert result is True
    assert result_shorthand is True


def test_or_filter_is_equal_when_sub_filters_are_equal():
    a_filter = Mock()
    a_filter.__eq__ = lambda _, x: x is a_filter
    b_filter = Mock()
    b_filter.__eq__ = lambda _, x: x is b_filter

    or_filter_1 = Or(a_filter, b_filter)
    or_filter_2 = Or(b_filter, a_filter)

    assert or_filter_1 == or_filter_2


def test_or_filter_is_not_equal_when_sub_filters_are_different():
    a_filter = Mock()
    a_filter.__eq__ = lambda _, x: x is a_filter
    b_filter = Mock()
    b_filter.__eq__ = lambda _, x: x is b_filter

    or_filter_1 = Or(a_filter, b_filter)
    or_filter_2 = Or(b_filter)
    or_filter_3 = Or(a_filter)

    assert or_filter_1 != or_filter_2
    assert or_filter_2 != or_filter_3


def test_or_filter_is_not_equal_when_other_is_not_a_filter():
    a_filter = Or(Mock())

    assert a_filter != 0


def test_or_filter_hash_is_hash_of_all_sub_filters():
    a_filter = Exists("a")
    b_filter = Exists("b")

    or_hash = hash(Or(a_filter, b_filter))

    assert or_hash == hash((Or, (a_filter, b_filter)))



from unittest.mock import Mock

import pytest

from src.event_processor.util import get_value_at_path


def test_get_value_at_path_gets_the_value_when_it_exists():
    path = "a.b.c"
    mock_value = Mock()

    result = get_value_at_path({"a": {"b": {"c": mock_value}}}, path)

    assert result is mock_value


def test_get_value_at_path_raises_for_an_invalid_path():
    path = "a.b.c"

    with pytest.raises(KeyError):
        get_value_at_path({"a": {"b": {"not-c": None}}}, path)



"""Dependency injection and management facilities."""
import inspect
from typing import Callable, Any, Optional, Dict, Tuple, List

try:
    from pydantic import BaseModel

    _has_pydantic = True
except ImportError:  # pragma: no cover
    _has_pydantic = False


class Event(dict):
    """Type to wrap a dict to be used as a dependency."""

    def __init__(self, dict_event: dict):
        super().__init__(dict_event)


class Depends:
    """Class to designate a dependency"""

    def __init__(self, callable_: Callable, cache: bool = True):
        """Create a dependency.

        :param callable_: The callable on which there is a dependency
        :param cache: Whether or not to cache the value resulting from this dependency
        """
        self.callable = callable_
        self.cache = cache

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.callable is other.callable and self.cache is other.cache:
                return True
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.callable, self.cache))


def call_with_injection(
    callable_: Callable, event: Optional[Event] = None, cache: Optional[dict] = None
) -> Optional[Any]:
    """Call a callable and inject required dependencies.

    Note that keyword args that have the same name as the parameter used for a dependency will be overwritten with the
    dependency's injected value.

    :param callable_: The callable to call
    :param event: The event for the current invocation
    :param cache: The dependency cache to use
    :return: The return value of the callable
    """
    value, _cacheable = resolve(Depends(callable_, cache=False), event=event, cache=cache)
    return value


def resolve(
    dependency: Depends, event: Optional[Event] = None, cache: Optional[dict] = None
) -> Tuple[Optional[Any], bool]:
    """Resolve a dependency into a value.

    The resulting values from dependencies are cached and re-used if a cache is supplied and the dependency itself
    does not explicitly state that it does not want to be cached. Also, any dependency that depends on another
    dependency where caching has been disabled will also not be cached (because the sub-value may change, which may in
    turn change the value of the current dependency).

    :param dependency: The dependency to resolve
    :param event: The event for the current invocation
    :param cache: The cache for previously resolved dependencies
    :return: The tuple (resolved_value, cacheable)
    :raises: pydantic.error_wrappers.ValidationError if the event cannot be parsed into a pydantic model
    """
    if cache and dependency in cache:
        return cache[dependency], True
    cacheable = dependency.cache

    resolved_dependencies: Dict[str, Any] = {}
    for arg_name in get_event_dependencies(dependency.callable):
        resolved_dependencies[arg_name] = event

    for arg_name, model in get_pydantic_dependencies(dependency.callable).items():
        resolved_dependencies[arg_name] = model.parse_obj(event)

    for arg_name, required_dependency in get_required_dependencies(dependency.callable).items():
        resolved_dependencies[arg_name], cacheable_dep = resolve(required_dependency, event=event, cache=cache)
        cacheable = cacheable and cacheable_dep

    value = dependency.callable(**resolved_dependencies)

    if cache is not None and cacheable:
        cache[dependency] = value

    return value, cacheable


def get_required_dependencies(callable_: Callable) -> Dict[str, Depends]:
    """Get the required dependencies for a callable.

    :param callable_: The callable for which to get dependencies
    :return: A mapping of callable argument names to dependencies
    """
    signature = inspect.signature(callable_)
    required_dependencies = {
        name: arg.default
        for name, arg in signature.parameters.items()
        if arg.default is not inspect.Parameter.empty and isinstance(arg.default, Depends)
    }
    return required_dependencies


def get_event_dependencies(callable_: Callable) -> List[str]:
    """Get the parameter names for event dependencies.

    :param callable_: The callable for which to get dependencies
    :return: A list of the parameters requiring the event
    """
    signature = inspect.signature(callable_)
    return [name for name, arg in signature.parameters.items() if arg.annotation is Event]


def get_pydantic_dependencies(callable_: Callable) -> Dict[str, "BaseModel"]:
    """Get the required models and their parameter names for a callable.

    :param callable_: The callable for which to get dependencies
    :return: A mapping of argument names to pydantic model types
    """
    if _has_pydantic:
        signature = inspect.signature(callable_)
        return {
            name: arg.annotation for name, arg in signature.parameters.items() if issubclass(arg.annotation, BaseModel)
        }
    else:
        return {}



import importlib
import sys
from typing import List
from unittest.mock import Mock, patch

from pydantic import BaseModel

from src.event_processor.dependencies import (
    Depends,
    resolve,
    get_required_dependencies,
    call_with_injection,
    Event,
    get_event_dependencies,
    get_pydantic_dependencies,
)


def test_depends_init_sets_values():
    def fn():
        pass

    result = Depends(fn, False)

    assert result.callable is fn
    assert result.cache is False


def test_depends_eq_compares_correctly():
    def fn():
        pass

    def other_fn():
        pass

    depends_fn = Depends(fn)
    depends_fn_again = Depends(fn)
    depends_fn_without_cache = Depends(fn, cache=False)
    depends_other_fn = Depends(other_fn)

    assert depends_fn == depends_fn_again
    assert depends_fn != depends_other_fn
    assert depends_fn != depends_fn_without_cache
    assert depends_fn != "not a depends"


def test_call_with_injection_calls_callable():
    callable_mock = Mock()

    call_with_injection(callable_mock)

    callable_mock.assert_called_once()


def test_call_with_injection_injects_dependencies():
    dependency = Mock()

    def fn(a: Mock = Depends(dependency)):
        a.method()

    call_with_injection(fn)

    dependency.return_value.method.assert_called_once()


def test_call_with_injection_injects_event_when_required():
    def fn(a: Event):
        return a

    event = Event({"a": "b"})

    result = call_with_injection(fn, event)

    assert result is event


def test_call_with_injection_returns_callable_value():
    callable_mock = Mock()

    result = call_with_injection(callable_mock)

    assert result is callable_mock.return_value


def test_resolve_calls_callable():
    mock_callable = Mock()
    dependency = Depends(mock_callable)

    resolve(dependency)

    mock_callable.assert_called_once()


def test_resolve_forwards_event_when_event_type_is_present_in_params():
    def fn(ev: Event):
        return ev

    dependency = Depends(fn)
    event = Event({"a": 0})

    resolved, _ = resolve(dependency, event)

    assert resolved is event


def test_resolve_injects_into_class_dependency():
    mock_service = Mock()
    event = Event({"a": 0})

    def dependency():
        return mock_service

    class Thing:
        def __init__(self, ev: Event, dep: Mock = Depends(dependency)):
            self.dep = dep
            self.ev = ev

    resolved, _ = resolve(Depends(Thing), event=event)

    assert resolved.ev == event
    assert resolved.dep is mock_service


def test_resolve_injects_pydantic_model_when_present_in_params():
    class Thing(BaseModel):
        id_: int
        items: List[str]

    def fn(thing: Thing):
        return thing

    resolved, _ = resolve(Depends(fn), event=Event({"id_": 1234, "items": ["a", "b"]}))

    assert resolved == Thing(id_=1234, items=["a", "b"])


def test_resolve_does_not_fail_when_the_event_contains_more_info_than_is_required():
    class Thing(BaseModel):
        id_: int
        items: List[str]

    def fn(thing: Thing):
        return thing

    resolved, _ = resolve(Depends(fn), event=Event({"id_": 1234, "items": ["a", "b"], "other": "stuff"}))

    assert resolved == Thing(id_=1234, items=["a", "b"])


def test_resolve_uses_cache_when_dependency_is_cached():
    dependency_mock = Mock()
    cache = {}

    dependency = Depends(dependency_mock)

    _resolved, _ = resolve(dependency, cache=cache)
    _resolved, _ = resolve(dependency, cache=cache)

    dependency_mock.assert_called_once()


def test_resolve_returns_callable_result():
    mock_callable = Mock()
    dependency = Depends(mock_callable)

    result, _ = resolve(dependency)

    assert result is mock_callable.return_value


def test_resolve_uses_cache_for_same_dependencies():
    mock_callable = Mock()
    dependency = Depends(mock_callable)
    cache = {}

    resolve(dependency, cache=cache)
    resolve(dependency, cache=cache)

    mock_callable.assert_called_once()


def test_resolve_does_not_use_cache_with_caching_disabled():
    mock_callable = Mock()
    dependency = Depends(mock_callable, cache=False)
    cache = {}

    resolve(dependency, cache=cache)
    resolve(dependency, cache=cache)

    assert len(mock_callable.mock_calls) == 2


def test_resolve_injects_required_dependencies_for_nested_dependencies():
    mock_callable = Mock()
    dependency = Depends(mock_callable, cache=False)

    def fn(a: Mock = dependency):
        return a

    nested_dependency = Depends(fn)
    result, _ = resolve(nested_dependency)

    assert result is mock_callable.return_value


def test_resolve_only_caches_values_that_require_other_cached_values():
    mock_callable = Mock()
    cached_dependency = Depends(mock_callable, cache=True)

    def fn(_=cached_dependency):
        pass

    no_cache_dependency = Depends(fn, cache=False)

    def no_cache_fn(_=no_cache_dependency):
        pass

    top_level_caching = Depends(no_cache_fn, cache=True)

    cache = {}
    resolve(top_level_caching, cache=cache)

    assert cached_dependency in cache
    assert no_cache_dependency not in cache
    assert top_level_caching not in cache  # Because it depends on a value that shouldn't be cached


def test_get_required_dependencies_returns_empty_list_for_no_dependencies():
    def fn():
        pass

    dependencies = get_required_dependencies(fn)

    assert dependencies == {}


def test_get_required_dependencies_returns_all_specified_dependencies():
    dependency = Depends(lambda: 0)
    other_dependency = Depends(lambda: 1)

    def fn(_a=dependency, _b=other_dependency):
        pass

    dependencies = get_required_dependencies(fn)

    assert dependencies == {"_a": dependency, "_b": other_dependency}


def test_get_event_dependencies_returns_dependant_param_names():
    def fn(_a: Event, _b: Event):
        pass

    param = get_event_dependencies(fn)

    assert param == ["_a", "_b"]


def test_get_event_dependencies_returns_empty_list_for_no_event_dependencies():
    callable_mock = Mock()

    param = get_event_dependencies(callable_mock)

    assert param == []


def test_get_pydantic_dependencies_returns_dependencies_when_they_are_specified():
    class Thing(BaseModel):
        pass

    def fn(_a: Thing):
        pass

    dependencies = get_pydantic_dependencies(fn)

    assert dependencies == {"_a": Thing}


def test_get_pydantic_dependencies_returns_no_dependencies_when_none_are_specified():
    def fn():
        pass

    dependencies = get_pydantic_dependencies(fn)

    assert dependencies == {}


@patch("src.event_processor.dependencies._has_pydantic", False)
def test_get_pydantic_dependencies_returns_none_even_when_present_if_pydantic_is_not_installed():
    def fn(_a: BaseModel):
        pass

    dependencies = get_pydantic_dependencies(fn)

    assert dependencies == {}



import pytest
from unittest.mock import Mock

from src.event_processor.exceptions import InvocationError
from src.event_processor.invocation_strategies import FirstMatch, AllMatches, NoMatches, NoMatchesStrict


def test_first_match_invokes_first_matching_processor():
    processor_a, processor_b = Mock(), Mock()

    FirstMatch.invoke([processor_a, processor_b])

    processor_a.assert_called_once()
    processor_b.assert_not_called()


def test_all_matches_invokes_all_matching_processors():
    processor_a, processor_b = Mock(), Mock()

    AllMatches.invoke([processor_a, processor_b])

    processor_a.assert_called_once()
    processor_b.assert_called_once()


def test_no_matches_invokes_no_matching_processors_on_multiple_matches():
    processor_a, processor_b = Mock(), Mock()

    NoMatches.invoke([processor_a, processor_b])

    processor_a.assert_not_called()
    processor_b.assert_not_called()


def test_no_matches_invokes_the_matching_processor_on_a_single_match():
    processor_a = Mock()

    NoMatches.invoke([processor_a])

    processor_a.assert_called_once()


def test_no_matches_strict_raises_when_multiple_processors_match():
    processor_a, processor_b = Mock(), Mock()

    with pytest.raises(InvocationError):
        NoMatchesStrict.invoke([processor_a, processor_b])


def test_no_matches_strict_invokes_the_matching_processor_on_a_single_match():
    processor_a = Mock()

    NoMatchesStrict.invoke([processor_a])

    processor_a.assert_called_once()



"""Contains the EventProcessor class."""
import inspect
from typing import Dict, Callable, Any, Tuple

from .dependencies import get_required_dependencies, get_event_dependencies, Event, Depends, get_pydantic_dependencies
from .exceptions import (
    FilterError,
    InvocationError,
)
from .filters import Filter
from .invocation_strategies import InvocationStrategies


class EventProcessor:
    """A self-contained event processor."""

    def __init__(self, invocation_strategy: InvocationStrategies = InvocationStrategies.FIRST_MATCH):
        self.processors: Dict[Tuple[Filter, int], Callable] = {}
        self.dependency_cache: Dict[Depends, Any] = {}
        self.invocation_strategy = invocation_strategy

    def add_subprocessor(self, subprocessor: "EventProcessor"):
        """Add a subprocessor to this event processor

        :param subprocessor: The other event processor to add
        """
        for filter_with_rank, processor in subprocessor.processors.items():
            if filter_with_rank in self.processors:
                raise FilterError(f"The filter '{filter_with_rank[0]}' is already handled by another processor")
            self.processors[filter_with_rank] = processor

    def processor(self, event_filter: Filter, rank: int = 0):
        """Register a new processor with the given filter and rank.

        :param event_filter: The filter for which to match events
        :param rank: This processor's rank (when there are multiple matches for a single event)
        """

        def decorate(fn):
            if not processor_params_are_valid(fn):
                raise FilterError(
                    f"The processor '{fn}' expects some invalid parameters "
                    f"(only dependencies, the event and pydantic models (if pydantic is installed) are allowed)"
                )
            if (event_filter, rank) in self.processors:
                raise FilterError(f"The filter '{event_filter}' ia already handled by another processor")

            self.processors[(event_filter, rank)] = fn
            return fn

        return decorate

    def invoke(self, event: Dict) -> Any:
        """Invoke the correct processor for an event.

        There may be multiple processors invoked, depending on the invocation strategy.

        :param event: The event to find a processor for
        :return: The return value of the processor
        """
        matching, highest_rank = [], float("-inf")
        for (filter_, rank), processor in self.processors.items():
            if filter_.matches(event):
                if rank > highest_rank:
                    matching, highest_rank = [processor], rank
                elif rank == highest_rank:
                    matching.append(processor)

        if matching:
            return self.invocation_strategy.value.invoke(matching, event=Event(event), cache=self.dependency_cache)
        else:
            raise InvocationError(f"No matching processor for the event '{event}'")


def processor_params_are_valid(processor: Callable) -> bool:
    """Verify that a processor's params are valid.

    :param processor: The processor to check
    :return: True if they are valid, False otherwise
    """
    dependencies = get_required_dependencies(processor)
    event_dependencies = get_event_dependencies(processor)
    pydantic_dependencies = get_pydantic_dependencies(processor)

    expected_dependency_count = len(dependencies) + len(event_dependencies) + len(pydantic_dependencies)

    return expected_dependency_count == len(inspect.signature(processor).parameters)



"""Contains many different filters to conveniently filter through events."""
from abc import ABC, abstractmethod
from typing import Any, Union, Callable

from .dependencies import call_with_injection, Event
from .exceptions import FilterError
from .util import get_value_at_path


class Filter(ABC):
    """Abstract filter to define the filter interface."""

    @abstractmethod
    def matches(self, event: dict) -> bool:
        """Test whether a given event matches an input event.

        :param event: The event to test
        :return: True if the event matches, False otherwise
        """

    @abstractmethod
    def __hash__(self):
        """Hash a filter for storage in a dict or set."""

    @abstractmethod
    def __eq__(self, other) -> bool:
        """Test if two filters are equal.

        :param other: The other filter to test
        :return: True if the filters are equal, False otherwise
        """

    def __ne__(self, other) -> bool:
        """Test if two filters are different.

        :param other: The other filter to test
        :return: True if the filters are different, False otherwise
        """

        return not (self == other)

    def __and__(self, other: "Filter") -> "Filter":
        """Combine two filters with a logical AND between them.

        :param other: The other filter to combine
        :return: A new filter representing the combination of the two
        """
        return And(self, other)

    def __or__(self, other: "Filter") -> "Filter":
        """Combine two filters with a logical OR between them.

        :param other: The other filter to combine
        :return: A new filter representing the combination of the two
        """
        return Or(self, other)


class Accept(Filter):
    """Accept any event (good for default processors)."""

    def matches(self, _event: dict) -> bool:
        return True

    def __hash__(self):
        return hash(self.__class__)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__)


class Exists(Filter):
    """Accept event where a given key exists."""

    def __init__(self, path: Any):
        self.path = path

    def matches(self, event: dict) -> bool:
        try:
            get_value_at_path(event, self.path)
        except KeyError:
            return False

        return True

    def __hash__(self):
        return hash((self.__class__, self.path))

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.path == other.path
        return False


class Eq(Filter):
    """Accept events where a given value is present at the given key."""

    def __init__(self, path: Any, value: Any):
        self.path = path
        self.value = value

    def matches(self, event: dict) -> bool:
        try:
            return self.value == get_value_at_path(event, self.path)
        except KeyError:
            return False

    def __hash__(self):
        return hash((self.__class__, (self.path, self.value)))

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.path == other.path and self.value == other.value
        return False


class NumCmp(Filter):
    """Accept events when the comparator returns True.

    If you use this processor, make sure that you don't use equal (and not identical) comparators for the same path.
    For example, don't use the same lambda in two different places. Instead, use a function, and pass a reference to
    that function. If you don't do that, the filters will effectively be different (even if they match the same thing),
    leading to perhaps unexpected results.
    """

    def __init__(self, path: Any, comparator: Callable[[float, float], bool], target: float):
        self.path = path
        self.comparator = comparator
        self.target = target

    def matches(self, event: dict) -> bool:
        try:
            found_value = get_value_at_path(event, self.path)
            float_value = float(found_value)
        except (KeyError, ValueError):
            return False

        return self.comparator(float_value, self.target)

    def __hash__(self):
        return hash((self.__class__, (self.path, self.comparator, self.target)))

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.path == other.path and self.comparator == other.comparator and self.target == other.target
        return False


class Lt(NumCmp):
    """Accept events where the value at the given path exists and is less than the specified value."""

    def __init__(self, path: Any, value: Union[int, float]):
        float_value = float(value)
        super().__init__(path, self._compare, float_value)

    @staticmethod
    def _compare(event_value: float, target_value: float) -> bool:
        return event_value < target_value


class Leq(NumCmp):
    """Accept events where the value at the given path exists is less than or equal to the specified value."""

    def __init__(self, path: Any, value: Union[int, float]):
        float_value = float(value)
        super().__init__(path, self._compare, float_value)

    @staticmethod
    def _compare(event_value: float, target_value: float) -> bool:
        return event_value <= target_value


class Gt(NumCmp):
    """Accept events where the value exists and is greater than the specified value."""

    def __init__(self, path: Any, value: Union[int, float]):
        float_value = float(value)
        super().__init__(path, self._compare, float_value)

    @staticmethod
    def _compare(event_value: float, target_value: float) -> bool:
        return event_value > target_value


class Geq(NumCmp):
    """Accept events where the value exists and is greater than or equal to the specified value."""

    def __init__(self, path: Any, value: Union[int, float]):
        float_value = float(value)
        super().__init__(path, self._compare, float_value)

    @staticmethod
    def _compare(event_value: float, target_value: float) -> bool:
        return event_value >= target_value


class Dyn(Filter):
    """Accept events based on a dynamic condition which is resolved by a callable.

    Note that the equality check with this filter is a bit special. The filter will only be equal if the resolver is
    the same (two functions with the same code are not equal, they need to be the same object in memory).

    This means that when using this filter, you should be careful to either re-use the same function as a resolver or
    make sure the functions have different behaviors. Otherwise, be sure to use processor ranking and invocation
    strategies.
    """

    def __init__(self, resolver: Callable):
        if not callable(resolver):
            raise FilterError("The resolver for a dyn filter must be callable")

        self.resolver = resolver

    def matches(self, event: dict) -> bool:
        if hasattr(self.resolver, "__name__") and self.resolver.__name__ == "<lambda>":
            return bool(self.resolver(event))

        return bool(call_with_injection(self.resolver, event=Event(event), cache={}))

    def __hash__(self):
        return hash(self.resolver)

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.resolver == other.resolver
        return False


class And(Filter):
    """Accept events that get accepted by all specified filters."""

    def __init__(self, *args: Filter):
        self.filters = args

    def matches(self, event: dict) -> bool:
        return all(filter_.matches(event) for filter_ in self.filters)

    def __hash__(self):
        return hash((self.__class__, self.filters))

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            self_in_other = all(filter_ in other.filters for filter_ in self.filters)
            other_in_self = all(filter_ in self.filters for filter_ in other.filters)
            return self_in_other and other_in_self
        return False


class Or(Filter):
    """Accept events that get accepted by at least one specified filter."""

    def __init__(self, *args: Filter):
        self.filters = args

    def matches(self, event: dict) -> bool:
        return any(filter_.matches(event) for filter_ in self.filters)

    def __hash__(self):
        return hash((self.__class__, self.filters))

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            self_in_other = all(filter_ in other.filters for filter_ in self.filters)
            other_in_self = all(filter_ in self.filters for filter_ in other.filters)
            return self_in_other and other_in_self
        return False



import os
from setuptools import setup, find_packages

# Meta info
project_root = os.path.dirname(os.path.realpath(__file__))
with open(f"{project_root}/VERSION", "r") as version_file:
    version = version_file.readline().strip()
with open(f"{project_root}/README.md", "r") as readme_file:
    readme = readme_file.read()

setup(
    name="event-processor",
    version=version,
    author="Nicolas Marier",
    author_email="software@nmarier.com",
    url="https://github.com/marier-nico/event-processor",
    project_urls={
        "Documentation": "https://event-processor.readthedocs.io/en/latest/",
        "Source": "https://github.com/marier-nico/event-processor",
        "Tracker": "https://github.com/marier-nico/event-processor/issues"
    },
    description="Pythonic event-processing library based on decorators",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="event decorators development",

    # Packages and depencies
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[],
    extras_require={"pydantic": ["pydantic ~= 1.8.2"]},
    package_data={"": ["VERSION"]},

    # Other configurations
    zip_safe=True,
    platforms="any",
)



"""Contains the different invocation strategies for calling processors."""
from abc import ABC
from enum import Enum
from typing import Dict, Optional, List, Callable, Any

from .exceptions import InvocationError
from .dependencies import call_with_injection, Event


class InvocationStrategy(ABC):
    """Class defining an abstract invocation strategy."""

    @staticmethod
    def invoke(matching: List[Callable], event: Optional[Event] = None, cache: Optional[Dict] = None) -> Any:
        """Invoke one or multiple matching processors."""


class FirstMatch(InvocationStrategy):
    """Strategy calling the first matching processor."""

    @staticmethod
    def invoke(matching: List[Callable], event: Optional[Event] = None, cache: Optional[Dict] = None) -> Any:
        return call_with_injection(matching[0], event=event, cache=cache)


class AllMatches(InvocationStrategy):
    """Strategy calling all matching processors."""

    @staticmethod
    def invoke(matching: List[Callable], event: Optional[Event] = None, cache: Optional[Dict] = None) -> Any:
        results = []
        for match in matching:
            results.append(call_with_injection(match, event=event, cache=cache))

        return tuple(results)


class NoMatches(InvocationStrategy):
    """Strategy not calling any matching processors."""

    @staticmethod
    def invoke(matching: List[Callable], event: Optional[Event] = None, cache: Optional[Dict] = None) -> Any:
        if len(matching) >= 2:
            return None

        return FirstMatch.invoke(matching, event=event, cache=cache)


class NoMatchesStrict(InvocationStrategy):
    """Strategy failing when there are multiple matching."""

    @staticmethod
    def invoke(matching: List[Callable], event: Optional[Event] = None, cache: Optional[Dict] = None) -> Any:
        if len(matching) >= 2:
            raise InvocationError("Multiple matching processors of the same rank")

        return FirstMatch.invoke(matching, event=event, cache=cache)


class InvocationStrategies(Enum):
    """Enumeration of available invocation strategies."""

    FIRST_MATCH = FirstMatch
    ALL_MATCHES = AllMatches
    NO_MATCHES = NoMatches
    NO_MATCHES_STRICT = NoMatchesStrict
